from typing import Literal

from pydantic import BaseModel


class BaseState(BaseModel):
    x_vel: float
    y_vel: float
    theta_deg_vel: float


class BaseCommand(BaseModel):
    x_vel: float
    y_vel: float
    theta_deg_vel: float


class ArmState(BaseModel):
    joint_angles: tuple[float, float, float, float, float]
    gripper_position: float | None = None


class ArmJointCommand(BaseModel):
    command_type: Literal["joint"] = "joint"
    joint_angles: tuple[float, float, float, float, float]
    gripper_position: float | None = None

    def __add__(self, other: "ArmJointCommand") -> "ArmJointCommand":
        return ArmJointCommand(
            command_type=self.command_type,
            joint_angles=tuple(
                a + b for a, b in zip(self.joint_angles, other.joint_angles)
            ),
            gripper_position=(
                self.gripper_position + other.gripper_position
                if (
                    self.gripper_position is not None
                    and other.gripper_position is not None
                )
                else self.gripper_position
            ),
        )

    def __sub__(self, other: "ArmJointCommand") -> "ArmJointCommand":
        return ArmJointCommand(
            command_type=self.command_type,
            joint_angles=tuple(
                a - b for a, b in zip(self.joint_angles, other.joint_angles)
            ),
            gripper_position=(
                self.gripper_position - other.gripper_position
                if (
                    self.gripper_position is not None
                    and other.gripper_position is not None
                )
                else self.gripper_position
            ),
        )

    def __mul__(self, other: float) -> "ArmJointCommand":
        return ArmJointCommand(
            command_type=self.command_type,
            joint_angles=tuple(a * other for a in self.joint_angles),
            gripper_position=self.gripper_position * other
            if self.gripper_position is not None
            else self.gripper_position
        )

    def __truediv__(self, other: float) -> "ArmJointCommand":
        return ArmJointCommand(
            command_type=self.command_type,
            joint_angles=tuple(a / other for a in self.joint_angles),
            gripper_position=self.gripper_position / other
            if self.gripper_position is not None
            else self.gripper_position
        )

    def clip(self, lo: "ArmJointCommand", hi: "ArmJointCommand") -> "ArmJointCommand":
        return ArmJointCommand(
            command_type=self.command_type,
            joint_angles=tuple(
                max(lo, min(a, hi))
                for a, lo, hi in zip(
                    self.joint_angles, lo.joint_angles, hi.joint_angles
                )
            ),
            gripper_position=(
                max(
                    lo.gripper_position, min(self.gripper_position, hi.gripper_position)
                )
                if self.gripper_position is not None
                else self.gripper_position
            ),
        )


class ArmEEPositionCommand(BaseModel):
    command_type: Literal["ee_position"] = "ee_position"
    xyz: tuple[float, float, float]
    gripper_position: float | None = None


class ArmEEInchingCommand(BaseModel):
    command_type: Literal["ee_inching"] = "ee_inching"
    delta_xyz: tuple[float, float, float]
    gripper_position: float | None = None


class LekiwiCommand(BaseModel):
    base_command: BaseCommand | None = None
    arm_command: ArmJointCommand | ArmEEPositionCommand | ArmEEInchingCommand | None = (
        None
    )



from dataclasses import dataclass
from typing import Callable, Generic, Optional, TypeVar


T = TypeVar("T")
ClipFn = Callable[[T, T, T], T]


def _zeros_like(x: T) -> T:
    return x - x


def _neg(x: T) -> T:
    return x * (-1)


def _scalar_clip(x: T, lo: T, hi: T) -> T:
    if x < lo:
        return lo
    if x > hi:
        return hi
    return x


def _default_clip(x: T, lo: T, hi: T) -> T:
    clip_m = getattr(x, "clip", None)
    if callable(clip_m):
        try:
            return clip_m(lo, hi)
        except TypeError:
            try:
                return clip_m(min=lo, max=hi)
            except TypeError:
                pass
    return _scalar_clip(x, lo, hi)


@dataclass
class AccelLimitedSmoother(Generic[T]):
    q: T
    v_max: T
    a_max: T
    dt: float
    tau_sec: float = 0.2
    v: Optional[T] = None
    clip_fn: Optional[ClipFn] = None

    def __post_init__(self) -> None:
        if self.v is None:
            self.v = _zeros_like(self.q)
        if self.clip_fn is None:
            self.clip_fn = _default_clip

    def step(self, q_target: T) -> tuple[T, T]:
        e = q_target - self.q
        v_goal = self.clip_fn(
            e / self.tau_sec,
            _neg(self.v_max),
            self.v_max,
        )
        a_dt = self.a_max * self.dt
        dv = v_goal - self.v
        dv = self.clip_fn(dv, _neg(a_dt), a_dt)
        self.v = self.v + dv
        self.q = self.q + self.v * self.dt
        return self.q, self.v



from pylekiwi.arm_controller import ArmController
from pylekiwi.base_controller import BaseController
from pylekiwi.models import ArmJointCommand, BaseCommand
from pylekiwi.settings import Settings

__all__ = [
    "ArmController",
    "ArmJointCommand",
    "BaseController",
    "BaseCommand",
    "Settings",
]



import numpy as np
from loguru import logger
from rustypot import Sts3215PyController

from pylekiwi.models import BaseCommand, BaseState
from pylekiwi.settings import Settings


class BaseController:
    LEFT_WHEEL_ID = 7
    BACK_WHEEL_ID = 8
    RIGHT_WHEEL_ID = 9

    # 角度φの定義（+x 前方を 0°、反時計回り）
    _PHI_DEG = np.array([60.0, 180.0, 300.0])  # [LEFT, BACK, RIGHT]

    def __init__(self, motor_controller: Sts3215PyController | Settings | None = None):
        if motor_controller is None:
            settings = (
                Settings()
                if not isinstance(motor_controller, Settings)
                else motor_controller
            )
            motor_controller = Sts3215PyController(
                serial_port=settings.serial_port,
                baudrate=settings.baudrate,
                timeout=settings.timeout,
            )
        elif isinstance(motor_controller, Settings):
            motor_controller = Sts3215PyController(
                serial_port=motor_controller.serial_port,
                baudrate=motor_controller.baudrate,
                timeout=motor_controller.timeout,
            )
        self.motor_controller = motor_controller

    def set_torque(self):
        for i in (self.LEFT_WHEEL_ID, self.BACK_WHEEL_ID, self.RIGHT_WHEEL_ID):
            self.motor_controller.write_torque_enable(i, True)

    def disable_torque(self):
        for i in (self.LEFT_WHEEL_ID, self.BACK_WHEEL_ID, self.RIGHT_WHEEL_ID):
            self.motor_controller.write_torque_enable(i, False)

    @staticmethod
    def _saturate_radps(w_radps: np.ndarray, max_radps: float = 10.0) -> np.ndarray:
        m = np.max(np.abs(w_radps))
        return w_radps if m <= max_radps else (w_radps * (max_radps / m))

    @classmethod
    def _body_to_wheel_radps(
        cls,
        vx: float,  # m/s
        vy: float,  # m/s
        vtheta_deg: float,  # deg/s
        wheel_radius: float = 0.05,
        base_radius: float = 0.125,
        max_radps: float = 10.0,  # 速度上限（rad/s）
    ) -> list[float]:
        # 角速度を rad/s に
        omega = np.deg2rad(vtheta_deg)
        phi = np.deg2rad(cls._PHI_DEG)  # [LEFT, BACK, RIGHT] の順

        # 標準形のキネマ行列（接線方向ベクトル）
        M = np.column_stack((-np.sin(phi), np.cos(phi), np.full(3, base_radius)))
        body = np.array([vx, vy, omega])

        # 車輪接線速度 -> 角速度（rad/s）
        wheel_radps = (M @ body) / wheel_radius

        # 安全のため物理上限で正規化
        wheel_radps = cls._saturate_radps(wheel_radps, max_radps=max_radps)
        return wheel_radps.tolist()

    @classmethod
    def _wheel_radps_to_body(
        cls,
        left_wheel_radps: float,
        back_wheel_radps: float,
        right_wheel_radps: float,
        wheel_radius: float = 0.05,
        base_radius: float = 0.125,
    ) -> BaseState:
        phi = np.deg2rad(cls._PHI_DEG)  # [LEFT, BACK, RIGHT]
        M = np.column_stack((-np.sin(phi), np.cos(phi), np.full(3, base_radius)))
        # 角速度 -> 接線速度（m/s）
        wheel_linear = wheel_radius * np.array(
            [left_wheel_radps, back_wheel_radps, right_wheel_radps]
        )
        # 逆運動学
        v = np.linalg.inv(M) @ wheel_linear
        vx, vy, omega = v
        return BaseState(x_vel=vx, y_vel=vy, theta_deg_vel=np.rad2deg(omega))

    def get_current_state(self) -> BaseState:
        wheel_radps = self.motor_controller.sync_read_present_speed(
            [self.LEFT_WHEEL_ID, self.BACK_WHEEL_ID, self.RIGHT_WHEEL_ID]
        )
        state = self._wheel_radps_to_body(wheel_radps)
        logger.debug(f"Base state: {state}")
        return state

    def send_action(self, action: BaseCommand):
        wheel_radps = self._body_to_wheel_radps(
            action.x_vel, action.y_vel, action.theta_deg_vel
        )
        self.motor_controller.sync_write_goal_speed(
            [self.LEFT_WHEEL_ID, self.BACK_WHEEL_ID, self.RIGHT_WHEEL_ID],
            wheel_radps,  # rad/s
        )

    def stop(self):
        self.motor_controller.sync_write_goal_speed(
            [self.LEFT_WHEEL_ID, self.BACK_WHEEL_ID, self.RIGHT_WHEEL_ID],
            [0.0, 0.0, 0.0],
        )



import cv2
import numpy as np


def encode_jpeg(frame: np.ndarray, quality: int = 80) -> bytes:
    ok, enc = cv2.imencode(
        ".jpg", frame, [int(cv2.IMWRITE_JPEG_QUALITY), int(quality)]
    )
    if not ok:
        raise RuntimeError("JPEG encode failed")
    return enc.tobytes()


class CameraController:
    def __init__(self, base_camera_id: int | None = None, arm_camera_id: int | None = None):
        self.cap_base = None
        self.cap_arm = None
        if base_camera_id is not None:
            self.cap_base = cv2.VideoCapture(base_camera_id)
        if arm_camera_id is not None:
            self.cap_arm = cv2.VideoCapture(arm_camera_id)

    def __del__(self):
        if self.cap_base is not None:
            self.cap_base.release()
        if self.cap_arm is not None:
            self.cap_arm.release()

    def get_base_frame(self) -> np.ndarray | None:
        if self.cap_base is None:
            return None
        ret, frame = self.cap_base.read()
        if not ret:
            return None
        return frame

    def get_arm_frame(self) -> np.ndarray | None:
        if self.cap_arm is None:
            return None
        ret, frame = self.cap_arm.read()
        if not ret:
            return None
        return frame



import os

import numpy as np
from loguru import logger
from rustypot import Sts3215PyController

from kinpy import build_serial_chain_from_mjcf

from pylekiwi.models import ArmJointCommand, ArmState
from pylekiwi.settings import Settings


_MODEL_FILE = os.path.join(os.path.dirname(__file__), "data/SO101/so101_new_calib.xml")


class ArmController:
    JOINT_IDS = (1, 2, 3, 4, 5)
    GRIPPER_ID = 6

    def __init__(self, motor_controller: Sts3215PyController | Settings | None = None):
        if motor_controller is None:
            settings = (
                Settings()
                if not isinstance(motor_controller, Settings)
                else motor_controller
            )
            motor_controller = Sts3215PyController(
                serial_port=settings.serial_port,
                baudrate=settings.baudrate,
                timeout=settings.timeout,
            )
        elif isinstance(motor_controller, Settings):
            motor_controller = Sts3215PyController(
                serial_port=motor_controller.serial_port,
                baudrate=motor_controller.baudrate,
                timeout=motor_controller.timeout,
            )
        self.motor_controller = motor_controller

        # self.chain = build_serial_chain_from_mjcf(
        #     open(_MODEL_FILE, "rb").read(),
        #     "gripper",
        #     model_dir=os.path.dirname(_MODEL_FILE),
        # )

    def set_torque(self):
        for i in self.JOINT_IDS:
            self.motor_controller.write_torque_enable(i, True)
        self.motor_controller.write_torque_enable(self.GRIPPER_ID, True)

    def disable_torque(self):
        for i in self.JOINT_IDS:
            self.motor_controller.write_torque_enable(i, False)
        self.motor_controller.write_torque_enable(self.GRIPPER_ID, False)

    def get_current_state(self) -> ArmState:
        all_ids = list(self.JOINT_IDS) + [self.GRIPPER_ID]
        joint_angles = self.motor_controller.sync_read_present_position(all_ids)
        gripper_position = joint_angles[-1]
        joint_angles = joint_angles[:-1]
        logger.debug(
            f"Joint angles: {joint_angles}, Gripper position: {gripper_position}"
        )
        return ArmState(
            joint_angles=tuple(joint_angles), gripper_position=gripper_position
        )

    def send_joint_action(self, action: ArmJointCommand):
        target_ids = list(self.JOINT_IDS)
        if action.gripper_position is not None:
            target_ids += [self.GRIPPER_ID]
        command = list(action.joint_angles)
        if action.gripper_position is not None:
            command += [action.gripper_position]
        self.motor_controller.sync_write_goal_position(target_ids, command)



import os
import time
import threading
from functools import partial

import numpy as np
import mesop as me
from pydantic import BaseModel
from rustypot import Sts3215PyController

from pylekiwi.models import BaseCommand, ArmJointCommand
from pylekiwi.base_controller import BaseController
from pylekiwi.arm_controller import ArmController
from pylekiwi.settings import Settings, constants
from pylekiwi.smoother import AccelLimitedSmoother


JOINT_LIMITS_DEG = [
    (-150.0, 150.0),  # J1
    (-120.0, 120.0),  # J2
    (-120.0, 120.0),  # J3
    (-150.0, 150.0),  # J4
    (-150.0, 150.0),  # J5
]
GRIPPER_LIMIT_DEG = (0.0, 60.0)
BORDER_LIGHT_SIDE = me.BorderSide(width=1, style="solid", color="#e5e7eb")
BORDER_LIGHT = me.Border(
    top=BORDER_LIGHT_SIDE,
    right=BORDER_LIGHT_SIDE,
    bottom=BORDER_LIGHT_SIDE,
    left=BORDER_LIGHT_SIDE,
)
PAD_N = lambda n: me.Padding(
    top=n,
    right=n,
    bottom=n,
    left=n,
)

# カードの共通スタイル
CARD = me.Style(
    display="flex",
    flex_direction="column",
    gap=16,
    padding=PAD_N(16),
    margin=me.Margin(top=12, bottom=12),
    border=BORDER_LIGHT,
    border_radius=16,
)


# 小さなバッジ風ラベル
def badge(text: str):
    with me.box(
        style=me.Style(
            display="inline-flex",
            align_items="center",
            padding=PAD_N(6),
            border_radius=9999,
            border=BORDER_LIGHT,
            gap=6,
        )
    ):
        me.text(text, type="caption")


class SharedBase(BaseModel):
    vx: float = 0.0  # m/s
    vy: float = 0.0  # m/s
    theta_deg: float = 0.0  # deg/s
    loop_enabled: bool = True
    armed: bool = False
    stop_requested: bool = True


class SharedArm(BaseModel):
    joints_deg: list[float] | None = None  # len=5
    gripper_deg: float = 0.0
    loop_enabled: bool = True
    armed: bool = False


shared_base = SharedBase()
shared_arm = SharedArm(joints_deg=[0.0] * 5, gripper_deg=0.0)

_controller: BaseController | None = None  # 台車
_arm: ArmController | None = None  # アーム

_base_loop_started = False
_arm_loop_started = False


def _start_base_loop():
    global _base_loop_started
    if _base_loop_started:
        return
    _base_loop_started = True

    def worker():
        last_sent_stop = True
        while True:
            try:
                if (
                    _controller
                    and shared_base.armed
                    and shared_base.loop_enabled
                    and not shared_base.stop_requested
                ):
                    cmd = BaseCommand(
                        x_vel=shared_base.vx,
                        y_vel=shared_base.vy,
                        theta_deg_vel=shared_base.theta_deg,
                    )
                    _controller.send_action(cmd)
                    last_sent_stop = False
                else:
                    if _controller and not last_sent_stop:
                        _controller.stop()
                        last_sent_stop = True
                time.sleep(0.05)  # 20Hz
            except Exception as e:
                print(f"[base-loop] {e}")
                time.sleep(0.2)

    threading.Thread(target=worker, daemon=True).start()


def _start_arm_loop():
    global _arm_loop_started
    if _arm_loop_started:
        return
    _arm_loop_started = True

    current_arm_state = _arm.get_current_state()
    current_arm_command = ArmJointCommand(
        joint_angles=current_arm_state.joint_angles,
        gripper_position=current_arm_state.gripper_position,
    )
    arm_smoother = AccelLimitedSmoother(
        q=current_arm_command,
        v_max=constants.JOINT_V_MAX,
        a_max=constants.JOINT_A_MAX,
        dt=constants.DT,
    )

    def worker():
        while True:
            try:
                time_start = time.time()
                if _arm and shared_arm.armed and shared_arm.loop_enabled:
                    joints_rad = tuple(np.deg2rad(shared_arm.joints_deg))
                    grip_rad = float(np.deg2rad(shared_arm.gripper_deg))
                    q, _ = arm_smoother.step(
                        ArmJointCommand(
                            joint_angles=joints_rad,
                            gripper_position=grip_rad,
                        )
                    )
                    _arm.send_joint_action(q)
                time.sleep(constants.DT - (time.time() - time_start))
            except Exception as e:
                print(f"[arm-loop] {e}")
                time.sleep(0.2)

    threading.Thread(target=worker, daemon=True).start()


@me.stateclass
class State:
    # ステータス
    connected: bool = False
    msg: str = ""

    # 台車パラメータ
    max_v: float = 0.25  # [m/s]
    max_omega_deg: float = 45.0  # [deg/s]
    base_armed: bool = False
    base_auto: bool = True

    # アームトグル
    arm_armed: bool = False
    arm_auto: bool = True

    # アーム関節値（deg表示）
    j1: float = 0.0
    j2: float = 0.0
    j3: float = 0.0
    j4: float = 0.0
    j5: float = 0.0
    grip: float = 0.0  # gripper


# ------------ INIT / LOAD ------------
def on_load(e: me.LoadEvent):
    me.set_theme_mode("system")
    global _controller, _arm

    try:
        if "LEKIWI_SERIAL_PORT" in os.environ:
            settings = Settings(serial_port=os.getenv("LEKIWI_SERIAL_PORT"))
        else:
            settings = Settings()
        motor_controller = Sts3215PyController(
            serial_port=settings.serial_port,
            baudrate=settings.baudrate,
            timeout=settings.timeout,
        )
        _controller = BaseController(motor_controller=motor_controller)
        _arm = ArmController(motor_controller=motor_controller)
        s = me.state(State)
        s.connected = True
        s.msg = "Controller connected"

        # 現在姿勢で初期化
        try:
            st = _arm.get_current_state()
            j_deg = [float(np.rad2deg(v)) for v in st.joint_angles]
            g_deg = (
                float(np.rad2deg(st.gripper_position))
                if st.gripper_position is not None
                else 0.0
            )
            _apply_arm_to_state(j_deg, g_deg)
        except Exception:
            pass

    except Exception as ex:
        s = me.state(State)
        s.connected = False
        s.msg = f"Init failed: {ex}"

    _start_base_loop()
    _start_arm_loop()


def _apply_arm_to_state(joints_deg: list[float], grip_deg: float):
    s = me.state(State)
    s.j1, s.j2, s.j3, s.j4, s.j5 = joints_deg
    s.grip = grip_deg
    shared_arm.joints_deg = joints_deg[:]
    shared_arm.gripper_deg = grip_deg


def _send_arm_once_from_state():
    if not _arm or not shared_arm.armed:
        return
    s = me.state(State)
    joints_deg = [s.j1, s.j2, s.j3, s.j4, s.j5]
    grip_deg = s.grip
    shared_arm.joints_deg = joints_deg[:]
    shared_arm.gripper_deg = grip_deg


def _stop_base():
    shared_base.vx = 0.0
    shared_base.vy = 0.0
    shared_base.theta_deg = 0.0
    shared_base.stop_requested = True
    if _controller:
        _controller.stop()


def _preset_speed(v: float, omega: float):
    s = me.state(State)
    s.max_v = v
    s.max_omega_deg = omega


@me.page(
    path="/",
    on_load=on_load,
    title="pylekiwi Web UI",
    security_policy=me.SecurityPolicy(
        allowed_script_srcs=[
            "'self'",
            "https://cdn.jsdelivr.net",
        ],
    ),
)
def page():
    s = me.state(State)

    # --- Header bar ---
    with me.box(
        style=me.Style(
            display="flex",
            justify_content="space-between",
            align_items="center",
            padding=PAD_N(16),
            border=BORDER_LIGHT,
            border_radius=16,
            margin=me.Margin(bottom=12),
        )
    ):
        with me.box(style=me.Style(display="flex", align_items="center", gap=12)):
            me.text("LeKiwi Teleop", type="headline-5")
            badge("CONNECTED" if s.connected else "DISCONNECTED")
        me.text(s.msg or "", type="body-1")
        with me.content_button(type="raised", on_click=on_estop, color="warn"):
            me.text("E-STOP")

    # --- Main grid: Base | Arm ---
    with me.box(
        style=me.Style(display="grid", gap=16, grid_template_columns="1fr 1fr")
    ):
        # ===== Base card =====
        with me.box(style=CARD):
            me.text("Base", type="headline-2")

            # Row: toggles + presets
            with me.box(
                style=me.Style(
                    display="flex", gap=12, align_items="center", flex_wrap="wrap"
                )
            ):
                me.slide_toggle(
                    label="Torque",
                    checked=s.base_armed,
                    on_change=on_toggle_base_arm,
                    color="accent",
                )
                me.slide_toggle(
                    label="Auto 20Hz",
                    checked=s.base_auto,
                    on_change=on_toggle_base_auto,
                )
                # Presets
                with me.content_button(
                    type="stroked", on_click=lambda e: _preset_speed(0.15, 60)
                ):
                    me.text("Slow")
                with me.content_button(
                    type="stroked", on_click=lambda e: _preset_speed(0.25, 120)
                ):
                    me.text("Normal")
                with me.content_button(
                    type="stroked", on_click=lambda e: _preset_speed(0.40, 180)
                ):
                    me.text("Fast")

            # Row: limits
            with me.box(
                style=me.Style(display="grid", grid_template_columns="1fr 1fr", gap=12)
            ):
                with me.box():
                    me.text(f"max_v = {s.max_v:.2f} m/s", type="headline-4")
                    me.slider(
                        min=0.05,
                        max=0.6,
                        step=0.01,
                        value=s.max_v,
                        on_value_change=on_change_max_v,
                        show_tick_marks=True,
                    )
                with me.box():
                    me.text(f"max_ω = {s.max_omega_deg:.0f} deg/s", type="headline-4")
                    me.slider(
                        min=10,
                        max=300,
                        step=1,
                        value=s.max_omega_deg,
                        on_value_change=on_change_max_omega,
                        show_tick_marks=True,
                    )

            # D-pad + rotation (3x3)
            me.text("D-pad", type="headline-4")
            with me.box(
                style=me.Style(
                    display="grid",
                    grid_template_columns="repeat(3, 110px)",
                    gap=10,
                    justify_content="start",
                )
            ):
                me.box()
                with me.content_button(type="flat", on_click=on_dpad_up):
                    me.text("↑")
                with me.content_button(type="flat", on_click=on_rot_left):
                    me.text("⟲")
                with me.content_button(type="flat", on_click=on_dpad_left):
                    me.text("←")
                with me.content_button(type="stroked", on_click=on_dpad_stop):
                    me.text("■ STOP")
                with me.content_button(type="flat", on_click=on_dpad_right):
                    me.text("→")
                me.box()
                with me.content_button(type="flat", on_click=on_dpad_down):
                    me.text("↓")
                with me.content_button(type="flat", on_click=on_rot_right):
                    me.text("⟳")

            # Current command snapshot
            me.text("Current command", type="headline-4")
            with me.box(style=me.Style(display="flex", gap=12, flex_wrap="wrap")):
                badge(f"vx {shared_base.vx:+.2f} m/s")
                badge(f"vy {shared_base.vy:+.2f} m/s")
                badge(f"ω  {shared_base.theta_deg:+.0f} deg/s")

        # ===== Arm card =====
        with me.box(style=CARD):
            me.text("Arm", type="headline-2")
            with me.box(
                style=me.Style(
                    display="flex", gap=12, flex_wrap="wrap", align_items="center"
                )
            ):
                me.slide_toggle(
                    label="Torque",
                    checked=s.arm_armed,
                    on_change=on_toggle_arm_arm,
                    color="accent",
                )
                me.slide_toggle(
                    label="Auto 10Hz",
                    checked=s.arm_auto,
                    on_change=on_toggle_arm_auto,
                    color="accent",
                )
                with me.content_button(type="stroked", on_click=on_arm_send_once):
                    me.text("Send")
                with me.content_button(type="stroked", on_click=on_arm_refresh):
                    me.text("Refresh")

            # Sliders (2 columns for joints + 1 row for gripper)
            with me.box(
                style=me.Style(display="grid", grid_template_columns="1fr 1fr", gap=12)
            ):
                for i, (lo, hi) in enumerate(JOINT_LIMITS_DEG):
                    val = getattr(s, f"j{i + 1}")
                    with me.box(
                        style=me.Style(display="flex", flex_direction="column", gap=6)
                    ):
                        me.text(f"J{i + 1}  {val:+.0f}°", type="headline-4")
                        me.slider(
                            min=lo,
                            max=hi,
                            step=1.0,
                            value=val,
                            on_value_change=partial(on_change_joint, i + 1),
                            style=me.Style(width="100%"),
                        )
            with me.box(
                style=me.Style(
                    display="flex",
                    gap=12,
                    align_items="center",
                    margin=me.Margin(top=8),
                )
            ):
                me.text(f"Gripper  {s.grip:.0f}°", type="headline-4")
                me.slider(
                    min=GRIPPER_LIMIT_DEG[0],
                    max=GRIPPER_LIMIT_DEG[1],
                    step=1.0,
                    value=s.grip,
                    on_value_change=on_change_gripper,
                    style=me.Style(width="100%"),
                )


def on_toggle_base_arm(e: me.SlideToggleChangeEvent):
    s = me.state(State)
    s.base_armed = not s.base_armed
    shared_base.armed = s.base_armed
    shared_base.stop_requested = not s.base_armed
    if _controller:
        if s.base_armed:
            _controller.set_torque()
        else:
            _stop_base()
            # トルクOFF（必要なら個別IDに対して）
            try:
                mc = _controller.motor_controller
                for wid in (
                    _controller.LEFT_WHEEL_ID,
                    _controller.BACK_WHEEL_ID,
                    _controller.RIGHT_WHEEL_ID,
                ):
                    mc.write_torque_enable(wid, False)
            except Exception:
                pass


def on_toggle_base_auto(e: me.SlideToggleChangeEvent):
    s = me.state(State)
    s.base_auto = not s.base_auto
    shared_base.loop_enabled = s.base_auto
    if not s.base_auto:
        shared_base.stop_requested = True


def on_change_max_v(e: me.SliderValueChangeEvent):
    me.state(State).max_v = e.value


def on_change_max_omega(e: me.SliderValueChangeEvent):
    me.state(State).max_omega_deg = e.value


def on_dpad_up(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = s.max_v, 0.0, 0.0
    shared_base.stop_requested = False


def on_dpad_down(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = -s.max_v, 0.0, 0.0
    shared_base.stop_requested = False


def on_dpad_left(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = 0.0, +s.max_v, 0.0
    shared_base.stop_requested = False


def on_dpad_right(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = 0.0, -s.max_v, 0.0
    shared_base.stop_requested = False


def on_rot_left(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = 0.0, 0.0, +s.max_omega_deg
    shared_base.stop_requested = False


def on_rot_right(e: me.ClickEvent):
    s = me.state(State)
    shared_base.vx, shared_base.vy, shared_base.theta_deg = 0.0, 0.0, -s.max_omega_deg
    shared_base.stop_requested = False


def on_dpad_stop(e: me.ClickEvent):
    _stop_base()


def on_estop(e: me.ClickEvent):
    # Base 停止＋トルクOFF
    s = me.state(State)
    s.base_armed = False
    shared_base.armed = False
    _stop_base()
    if _controller:
        try:
            _controller.disable_torque()
        except Exception:
            pass
    # Arm もOFF
    s.arm_armed = False
    shared_arm.armed = False
    if _arm:
        try:
            _arm.disable_torque()
        except Exception:
            pass


def on_toggle_arm_arm(e: me.SlideToggleChangeEvent):
    s = me.state(State)
    s.arm_armed = not s.arm_armed
    shared_arm.armed = s.arm_armed
    if _arm:
        try:
            if s.arm_armed:
                _arm.set_torque()
            else:
                _arm.disable_torque()
        except Exception as ex:
            print(f"[arm-torque] {ex}")


def on_toggle_arm_auto(e: me.SlideToggleChangeEvent):
    s = me.state(State)
    s.arm_auto = not s.arm_auto
    shared_arm.loop_enabled = s.arm_auto
    if not s.arm_auto:
        shared_arm.armed = False


def on_arm_send_once(e: me.ClickEvent):
    _send_arm_once_from_state()


def on_arm_refresh(e: me.ClickEvent):
    if not _arm:
        return
    try:
        st = _arm.get_current_state()
        j_deg = [float(np.rad2deg(v)) for v in st.joint_angles]
        g_deg = (
            float(np.rad2deg(st.gripper_position))
            if st.gripper_position is not None
            else 0.0
        )
        _apply_arm_to_state(j_deg, g_deg)
    except Exception as ex:
        print(f"[arm-refresh] {ex}")


def on_change_joint(idx: int, e: me.SliderValueChangeEvent):
    s = me.state(State)
    setattr(s, f"j{idx}", e.value)
    if shared_arm.armed:
        _send_arm_once_from_state()


def on_change_gripper(e: me.SliderValueChangeEvent):
    s = me.state(State)
    s.grip = e.value
    if shared_arm.armed:
        _send_arm_once_from_state()



from typer import Typer

from loguru import logger

from pylekiwi.nodes import LeaderControllerNode
from pylekiwi.settings import Settings


app = Typer(help="Launch the leader controller", invoke_without_command=True)


@app.callback()
def leader(serial_port: str = "/dev/ttyACM0"):
    logger.info("Starting leader controller node")
    logger.info(f"Serial port: {serial_port}")
    logger.info("Base command is from the following keys:")
    logger.info("| w  | forward        |")
    logger.info("| s  | backward       |")
    logger.info("| a  | left           |")
    logger.info("| d  | right          |")
    logger.info("| -> | left rotation  |")
    logger.info("| <- | right rotation |")
    leader_node = LeaderControllerNode(settings=Settings(serial_port=serial_port))
    leader_node.run()


if __name__ == "__main__":
    app()



from typer import Typer

from loguru import logger

from pylekiwi.nodes import HostControllerNode
from pylekiwi.settings import Settings


app = Typer(help="Launch the host controller", invoke_without_command=True)

@app.callback()
def host(serial_port: str = "/dev/ttyACM0"):
    logger.info("Starting host controller node")
    logger.info(f"Serial port: {serial_port}")
    host_controller_node = HostControllerNode(Settings(serial_port=serial_port))
    host_controller_node.run()


if __name__ == "__main__":
    app()



from pynput import keyboard

from pylekiwi.models import BaseCommand


class KeyListener:
    def __init__(self, vel: float = 0.25, omega_deg: float = 45.0):
        self.current_command = BaseCommand(x_vel=0, y_vel=0, theta_deg_vel=0)
        self.vel = vel
        self.omega_deg = omega_deg

    def on_key_press(self, key: keyboard.Key) -> None:
        if hasattr(key, "char"):
            if key.char == "w":
                self.current_command = BaseCommand(x_vel=self.vel, y_vel=0, theta_deg_vel=0)
            elif key.char == "s":
                self.current_command = BaseCommand(x_vel=-self.vel, y_vel=0, theta_deg_vel=0)
            elif key.char == "a":
                self.current_command = BaseCommand(x_vel=0, y_vel=self.vel, theta_deg_vel=0)
            elif key.char == "d":
                self.current_command = BaseCommand(x_vel=0, y_vel=-self.vel, theta_deg_vel=0)
        elif key == keyboard.Key.left:
            self.current_command = BaseCommand(x_vel=0, y_vel=0, theta_deg_vel=self.omega_deg)
        elif key == keyboard.Key.right:
            self.current_command = BaseCommand(x_vel=0, y_vel=0, theta_deg_vel=-self.omega_deg)
    
    def on_key_release(
        self,
        key: keyboard.Key,
    ) -> None:
        self.current_command = BaseCommand(x_vel=0, y_vel=0, theta_deg_vel=0)



import time

import cv2
import numpy as np
import zenoh
try:
    import rerun as rr
except ImportError:
    rr = None
from collections import deque
from loguru import logger
from rustypot import Sts3215PyController

from pylekiwi.arm_controller import ArmController
from pylekiwi.base_controller import BaseController
from pylekiwi.camera_controller import CameraController, encode_jpeg
from pylekiwi.models import ArmJointCommand, BaseCommand, LekiwiCommand
from pylekiwi.settings import Settings, constants
from pylekiwi.smoother import AccelLimitedSmoother


class HostControllerNode:
    """Host controller node that receives commands and sends them to the base and arm controllers.
    """

    def __init__(self, settings: Settings | None = None):
        settings = settings or Settings()
        motor_controller = Sts3215PyController(
            serial_port=settings.serial_port,
            baudrate=settings.baudrate,
            timeout=settings.timeout,
        )
        self._base_controller = BaseController(motor_controller=motor_controller)
        self._arm_controller = ArmController(motor_controller=motor_controller)
        self._camera_controller = CameraController(
            base_camera_id=settings.base_camera_id,
            arm_camera_id=settings.arm_camera_id,
        )
        self._target_arm_command: ArmJointCommand | None = None
        self._dt = constants.DT

    def _listener(self, msg: zenoh.Sample) -> zenoh.Reply:
        command: LekiwiCommand = LekiwiCommand.model_validate_json(msg.payload.to_string())
        logger.debug(f"Received command: {command}")
        if command.base_command is not None:
            self._base_controller.send_action(command.base_command)
        if (
            command.arm_command is not None
            and command.arm_command.command_type == "joint"
        ):
            self._target_arm_command = command.arm_command

    def run(self):
        with zenoh.open(zenoh.Config()) as session:
            sub = session.declare_subscriber(constants.COMMAND_KEY, self._listener)
            pub_base_cam = session.declare_publisher(constants.BASE_CAMERA_KEY)
            pub_arm_cam = session.declare_publisher(constants.ARM_CAMERA_KEY)
            try:
                current_arm_state = self._arm_controller.get_current_state()
                current_arm_command = ArmJointCommand(
                    joint_angles=current_arm_state.joint_angles,
                    gripper_position=current_arm_state.gripper_position,
                )
                self._arm_smoother = AccelLimitedSmoother(
                    q=current_arm_command,
                    v_max=constants.JOINT_V_MAX,
                    a_max=constants.JOINT_A_MAX,
                    dt=self._dt,
                )
                self._target_arm_command = current_arm_command
            except Exception as e:
                logger.error(f"Error initializing arm smoother: {e}")
                sub.undeclare()
                return
            logger.info("Starting host controller node...")
            try:
                while True:
                    start_time = time.time()
                    if self._target_arm_command is not None:
                        q, _ = self._arm_smoother.step(self._target_arm_command)
                        self._arm_controller.send_joint_action(q)
                    # Publish camera frames
                    base_frame = self._camera_controller.get_base_frame()
                    arm_frame = self._camera_controller.get_arm_frame()
                    if base_frame is not None:
                        pub_base_cam.put(encode_jpeg(base_frame))
                    if arm_frame is not None:
                        pub_arm_cam.put(encode_jpeg(arm_frame))
                    time.sleep(max(0, self._dt - (time.time() - start_time)))
            except KeyboardInterrupt:
                pass
            finally:
                sub.undeclare()


class ClientControllerNode:
    """Controller node that publishes commands to the host node.
    """

    def __init__(self):
        self.session = zenoh.open(zenoh.Config())
        self.publisher = self.session.declare_publisher(constants.COMMAND_KEY)

    def send_command(self, command: LekiwiCommand):
        self.publisher.put(command.model_dump_json())

    def send_base_command(self, command: BaseCommand):
        self.send_command(LekiwiCommand(base_command=command))

    def send_arm_joint_command(self, command: ArmJointCommand):
        self.send_command(LekiwiCommand(arm_command=command))


class ClientControllerWithCameraNode(ClientControllerNode):
    """Controller node that publishes commands to the host node and receives camera frames.
    """

    def __init__(self, settings: Settings):
        super().__init__()
        self.settings = settings
        self.sub_base_cam = self.session.declare_subscriber(constants.BASE_CAMERA_KEY, self._listener_base_cam)
        self.sub_arm_cam = self.session.declare_subscriber(constants.ARM_CAMERA_KEY, self._listener_arm_cam)
        self.base_frame_queue = deque(maxlen=5)
        self.arm_frame_queue = deque(maxlen=5)
        if rr is not None and settings.view_camera:
            rr.init("lekiwi_client_camera", spawn=settings.rerun_spawn)

    def _listener_base_cam(self, msg: zenoh.Sample) -> zenoh.Reply:
        binary_data = bytes(msg.payload)
        image = cv2.imdecode(np.frombuffer(binary_data, dtype=np.uint8), cv2.IMREAD_COLOR)
        self.base_frame_queue.append(image)

    def _listener_arm_cam(self, msg: zenoh.Sample) -> zenoh.Reply:
        binary_data = bytes(msg.payload)
        image = cv2.imdecode(np.frombuffer(binary_data, dtype=np.uint8), cv2.IMREAD_COLOR)
        self.arm_frame_queue.append(image)

    def get_base_frame(self) -> np.ndarray | None:
        return self.base_frame_queue[-1] if len(self.base_frame_queue) > 0 else None

    def get_arm_frame(self) -> np.ndarray | None:
        return self.arm_frame_queue[-1] if len(self.arm_frame_queue) > 0 else None

    def view_camera(self):
        if (
            rr is not None
            and self.settings.view_camera
            and len(self.base_frame_queue) > 0
            and len(self.arm_frame_queue) > 0
        ):
            rr.log("base_camera", rr.Image(self.base_frame_queue[-1][..., ::-1]))
            rr.log("arm_camera", rr.Image(self.arm_frame_queue[-1][..., ::-1]))


class LeaderControllerNode(ClientControllerWithCameraNode):
    """Leader controller node that publishes commands to the host node.
    Arm commands are based on the current leader's arm state.
    Base commands are from the keyboard.
    """

    def __init__(self, settings: Settings | None = None):
        settings = settings or Settings()
        super().__init__(settings=settings)
        try:
            motor_controller = Sts3215PyController(
                serial_port=settings.serial_port,
                baudrate=settings.baudrate,
                timeout=settings.timeout,
            )
            self.arm_controller = ArmController(motor_controller=motor_controller)
        except OSError as e:
            logger.error(f"Error initializing arm controller: {e}")
            self.arm_controller = None

        from pylekiwi.key_listener import KeyListener
        self.key_listener = KeyListener()

    def send_leader_command(self, base_command: BaseCommand | None = None):
        if self.arm_controller is not None:
            arm_state = self.arm_controller.get_current_state()
            arm_command = ArmJointCommand(
                joint_angles=arm_state.joint_angles,
                gripper_position=arm_state.gripper_position,
            )
            if base_command is not None:
                self.send_command(LekiwiCommand(base_command=base_command, arm_command=arm_command))
            else:
                self.send_arm_joint_command(arm_command)
        elif base_command is not None:
            self.send_base_command(base_command)

    def run(self):
        from pynput import keyboard

        with keyboard.Listener(
            on_press=self.key_listener.on_key_press,
            on_release=self.key_listener.on_key_release,
        ):
            while True:
                start_time = time.time()
                self.view_camera()
                self.send_leader_command(base_command=self.key_listener.current_command)
                time.sleep(max(0, constants.DT - (time.time() - start_time)))



from pydantic import BaseModel

from pylekiwi.models import ArmJointCommand


class Settings(BaseModel):
    serial_port: str = "/dev/ttyACM0"
    baudrate: int = 1000000
    timeout: float = 0.5
    base_camera_id: int | None = 0
    arm_camera_id: int | None = 2
    view_camera: bool = True
    rerun_spawn: bool = True


class Constants(BaseModel):
    DT: float = 0.01
    JOINT_V_MAX: ArmJointCommand = ArmJointCommand(
        joint_angles=(80.0, 80.0, 80.0, 80.0, 80.0),
        gripper_position=80.0,
    )
    JOINT_A_MAX: ArmJointCommand = ArmJointCommand(
        joint_angles=(600.0, 600.0, 600.0, 600.0, 600.0),
        gripper_position=600.0,
    )
    COMMAND_KEY: str = "lekiwi/command"
    BASE_CAMERA_KEY: str = "lekiwi/camera/base"
    ARM_CAMERA_KEY: str = "lekiwi/camera/arm"


constants = Constants()



from typer import Typer

from pylekiwi.commands import launch_host_controller, launch_leader, launch_web_ui


app = Typer(help="pylekiwi: Python package for controlling the LeKiwi robot", no_args_is_help=True)


app.add_typer(launch_host_controller.app, name="host")
app.add_typer(launch_leader.app, name="leader")
app.add_typer(launch_web_ui.app, name="webui")



from typer import Typer
import subprocess


app = Typer(help="Launch the web UI", invoke_without_command=True)

@app.callback()
def web_ui(port: int = 8080, serial_port: str = "/dev/ttyACM0"):
    proc = subprocess.run(
        [
            "gunicorn",
            "--bind",
            f"0.0.0.0:{port}",
            "--env",
            f"LEKIWI_SERIAL_PORT={serial_port}",
            "pylekiwi.commands.web_ui:me",
        ]
    )
    return proc.returncode
